"""
data container for option chain
data is provided as:
1 arrays of ttms, forwards, discounts
2 lists of arrays with strikes, optiom types and bid / ask prices and vols
"""

from __future__ import annotations

import numpy as np
import pandas as pd
from dataclasses import dataclass
from typing import Tuple, Optional
from numba.typed import List

import stochvolmodels.pricers.analytic.bsm as bsm
import stochvolmodels.pricers.analytic.bachelier as bachel
from stochvolmodels.utils.var_swap_pricer import compute_var_swap_strike
from stochvolmodels.pricers.factor_hjm.rate_core import get_default_swap_term_structure, swap_rate


@dataclass
class OptionSlice:
    """
    container for slice data
    """
    ttm: float
    forward: float
    strikes: np.ndarray
    optiontypes: np.ndarray
    id: str
    discfactor: float = None  # discount factors
    discount_rate: float = None  # discount rates
    bid_ivs: Optional[np.ndarray] = None
    ask_ivs: Optional[np.ndarray] = None
    bid_prices: Optional[np.ndarray] = None
    ask_prices: Optional[np.ndarray] = None

    def __post_init__(self):
        """
        to do: check dimension aligmnent
        make consistent discfactors
        """
        if self.discfactor is not None:
            self.discount_rate = - np.log(self.discfactor) / self.ttm
        elif self.discount_rate is not None:
            self.discfactor = np.exp(-self.discount_rate * self.ttm)
        else:  # use zeros
            self.discfactor = 1.0
            self.discount_rate = 0.0


@dataclass
class OptionChain:
    """
    container for chain data
    note we do not use chain as list of slices here
    for extensive use of numba we use List[np.ndarray] with per slice data
    """
    ttms: np.ndarray
    forwards: np.ndarray
    strikes_ttms: List[np.ndarray]
    optiontypes_ttms: List[np.ndarray]
    ids: Optional[np.ndarray]  # slice_t names
    discfactors: Optional[np.ndarray] = None  # discount factors
    discount_rates: Optional[np.ndarray] = None  # discount rates
    ticker: Optional[str] = None  # associated ticker
    bid_ivs: Optional[List[np.ndarray]] = None
    ask_ivs: Optional[List[np.ndarray]] = None
    bid_prices: Optional[List[np.ndarray]] = None
    ask_prices: Optional[List[np.ndarray]] = None
    forwards0: Optional[np.ndarray] = None  # when we need to normalize options price

    def __post_init__(self):
        """
        to do: check dimension aligmnent
        make consistent discfactors
        """
        if self.discfactors is not None:
            self.discount_rates = - np.log(self.discfactors) / self.ttms
        elif self.discount_rates is not None:
            self.discfactors = np.exp(-self.discount_rates * self.ttms)
        else:  # use zeros
            self.discfactors = np.ones_like(self.ttms)
            self.discount_rates = np.zeros_like(self.ttms)

    def print(self) -> None:
        this = dict(ttms=self.ttms,
                    forwards=self.forwards,
                    strikes_ttms=self.strikes_ttms,
                    optiontypes_ttms=self.optiontypes_ttms,
                    ids=self.ids,
                    bid_ivs=self.bid_ivs,
                    ask_ivs=self.ask_ivs)
        for k, v in this.items():
            print(f"{k}:\n{v}")

    @classmethod
    def slice_to_chain(cls,
                       ttm: float,
                       forward: float,
                       strikes: np.ndarray,
                       optiontypes: np.ndarray,
                       discfactor: float = 1.0,
                       id: Optional[str] = None
                       ) -> OptionChain:

        return cls(ttms=np.array([ttm]),
                   forwards=np.array([forward]),
                   strikes_ttms=(strikes,),
                   optiontypes_ttms=(optiontypes,),
                   discfactors=np.array([discfactor]),
                   ids=np.array([id]) if id is not None else np.array([f"{ttm:0.2f}"]))

    def get_mid_vols(self) -> List[np.ndarray]:
        if self.bid_ivs is not None and self.ask_ivs is not None:
            return List(0.5 * (bid_iv + ask_iv) for bid_iv, ask_iv in zip(self.bid_ivs, self.ask_ivs))
        else:
            return None

    def get_chain_deltas(self) -> List[np.ndarray]:
        deltas_ttms = bsm.compute_bsm_vanilla_deltas_ttms(ttms=self.ttms,
                                                          forwards=self.forwards,
                                                          strikes_ttms=self.strikes_ttms,
                                                          optiontypes_ttms=self.optiontypes_ttms,
                                                          vols_ttms=self.get_mid_vols())
        return deltas_ttms

    def get_chain_vegas(self, is_unit_ttm_vega: bool = False) -> List[np.ndarray]:
        if is_unit_ttm_vega:
            ttms = np.ones_like(self.ttms)
        else:
            ttms = self.ttms
        vegas_ttms = bsm.compute_bsm_vegas_ttms(ttms=ttms,
                                                forwards=self.forwards,
                                                strikes_ttms=self.strikes_ttms,
                                                optiontypes_ttms=self.optiontypes_ttms,
                                                vols_ttms=self.get_mid_vols())
        return vegas_ttms

    def get_chain_atm_vols(self) -> np.ndarray:
        atm_vols = np.zeros(len(self.ttms))
        for idx, (forward, strikes_ttm, y) in enumerate(zip(self.forwards, self.strikes_ttms, self.get_mid_vols())):
            atm_vols[idx] = np.interp(x=forward, xp=strikes_ttm, fp=y)
        return atm_vols

    def get_chain_skews(self, delta: float = 0.25) -> np.ndarray:
        skews = np.zeros(len(self.ttms))
        deltas_ttms = self.get_chain_deltas()
        for idx, (deltas, vols) in enumerate(zip(deltas_ttms, self.get_mid_vols())):
            dput = np.interp(x=-delta, xp=deltas, fp=vols)
            d50 = np.interp(x=0.5, xp=deltas, fp=vols)
            dcall= np.interp(x=delta, xp=deltas, fp=vols)
            skews[idx] = (dput - dcall)/d50
        return skews

    def get_chain_data_as_xy(self) -> Tuple[List[np.ndarray], np.ndarray]:
        """
        these data are needed for to pass x and y for model calibrations
        """
        mid_vols = List(0.5 * (bid_iv + ask_iv) for bid_iv, ask_iv in zip(self.bid_ivs, self.ask_ivs))
        x = (self.ttms, self.forwards, self.discfactors, self.strikes_ttms, self.optiontypes_ttms)
        y = mid_vols
        return x, y

    def compute_model_ivols_from_chain_data(self,
                                            model_prices: List[np.ndarray],
                                            forwards: np.ndarray = None
                                            ) -> List[np.ndarray]:
        if forwards is None:
            forwards = self.forwards

        model_ivols = bsm.infer_bsm_ivols_from_model_chain_prices(ttms=self.ttms,
                                                                  forwards=forwards,
                                                                  discfactors=self.discfactors,
                                                                  strikes_ttms=self.strikes_ttms,
                                                                  optiontypes_ttms=self.optiontypes_ttms,
                                                                  model_prices_ttms=model_prices)
        return model_ivols

    @classmethod
    def to_forward_normalised_strikes(cls, obj: OptionChain) -> OptionChain:
        """
        strike and prices to normalized by forwards
        """
        new_strikes_ttms = List()
        for strikes_ttm, forward in zip(obj.strikes_ttms, obj.forwards):
            new_strikes = strikes_ttm / forward
            new_strikes_ttms.append(new_strikes)

        return cls(ttms=obj.ttms,
                   forwards=np.ones_like(obj.forwards),
                   strikes_ttms=new_strikes_ttms,
                   optiontypes_ttms=obj.optiontypes_ttms,
                   discfactors=obj.discfactors,
                   ticker=obj.ticker,
                   ids=obj.ids,
                   bid_ivs=obj.bid_ivs, ask_ivs=obj.ask_ivs,
                   forwards0=obj.forwards)

    @classmethod
    def to_uniform_strikes(cls, obj: OptionChain, num_strikes: int = 21) -> OptionChain:
        """
        in some situations (like model price display) we want to get a uniform grid corresponding to the chain
        bid_ivs and ask_ivs will be set to none
        """
        new_strikes_ttms = List()
        new_optiontypes_ttms = List()
        for strikes_ttm, forward in zip(obj.strikes_ttms, obj.forwards):
            new_strikes = np.linspace(strikes_ttm[0], strikes_ttm[-1], num_strikes)
            new_strikes_ttms.append(new_strikes)
            new_optiontypes_ttms.append(np.where(new_strikes >= forward, 'C', 'P'))

        return cls(ttms=obj.ttms, forwards=obj.forwards, strikes_ttms=new_strikes_ttms,
                   optiontypes_ttms=new_optiontypes_ttms, discfactors=obj.discfactors,
                   ticker=obj.ticker,
                   ids=obj.ids,
                   bid_ivs=None, ask_ivs=None)

    def get_slice(self, id: str) -> OptionSlice:
        idx = list(self.ids).index(id)
        option_slice = OptionSlice(id=self.ids[idx],
                                   ttm=self.ttms[idx],
                                   forward=self.forwards[idx],
                                   strikes=self.strikes_ttms[idx],
                                   optiontypes=self.optiontypes_ttms[idx],
                                   discfactor=self.discfactors[idx],
                                   bid_ivs=None if self.bid_ivs is None else self.bid_ivs[idx],
                                   ask_ivs=None if self.ask_ivs is None else self.ask_ivs[idx],
                                   bid_prices=None if self.bid_prices is None else self.bid_prices[idx],
                                   ask_prices=None if self.ask_prices is None else self.ask_prices[idx])
        return option_slice

    def get_slice_varswap_strikes(self, floor_with_atm_vols: bool = True) -> pd.Series:
        varswap_strikes = np.zeros_like(self.ttms)
        vols_ttms = self.get_mid_vols()
        for idx, ttm in enumerate(self.ttms):
            mid_prices = bsm.compute_bsm_vanilla_slice_prices(ttm=ttm,
                                                              forward=self.forwards[idx],
                                                              strikes=self.strikes_ttms[idx],
                                                              vols=vols_ttms[idx],
                                                              optiontypes=self.optiontypes_ttms[idx])
            strikes = self.strikes_ttms[idx]
            puts_cond = self.optiontypes_ttms[idx] == 'P'
            puts = pd.Series(mid_prices[puts_cond], index=strikes[puts_cond])
            calls = pd.Series(mid_prices[puts_cond == False], index=strikes[puts_cond == False])
            varswap_strikes[idx] = compute_var_swap_strike(puts=puts, calls=calls, forward=self.forwards[idx], ttm=ttm)

        if floor_with_atm_vols:
            varswap_strikes = np.maximum(self.get_chain_atm_vols(), varswap_strikes)

        return pd.Series(varswap_strikes, index=self.ttms)

    @classmethod
    def get_slices_as_chain(cls, option_chain: OptionChain, ids: List[str]) -> OptionChain:
        """
        return a subset of chain for given ids
        """
        if len(ids) == 1:
            idx = list(option_chain.ids).index(ids[0])  # get location of ids
            option_chain = cls(ids=ids,
                               ttms=np.array([option_chain.ttms[idx]]),
                               ticker=option_chain.ticker,
                               forwards=np.array([option_chain.forwards[idx]]),
                               strikes_ttms=List([option_chain.strikes_ttms[idx]]),
                               optiontypes_ttms=List([option_chain.optiontypes_ttms[idx]]),
                               discfactors=np.array([option_chain.discfactors[idx]]),
                               bid_ivs=None if option_chain.bid_ivs is None else List([option_chain.bid_ivs[idx]]),
                               ask_ivs=None if option_chain.ask_ivs is None else List([option_chain.ask_ivs[idx]]),
                               bid_prices=None if option_chain.bid_prices is None else List([option_chain.bid_prices[idx]]),
                               ask_prices=None if option_chain.ask_prices is None else List([option_chain.ask_prices[idx]]))
        else:
            indices = np.isin(option_chain.ids, ids).nonzero()[0]
            option_chain = cls(ids=ids,
                               ttms=option_chain.ttms[indices],
                               ticker=option_chain.ticker,
                               forwards=option_chain.forwards[indices],
                               strikes_ttms=List(option_chain.strikes_ttms[idx] for idx in indices),
                               optiontypes_ttms=List(option_chain.optiontypes_ttms[idx] for idx in indices),
                               discfactors=option_chain.discfactors[indices],
                               bid_ivs=None if option_chain.bid_ivs is None else List(option_chain.bid_ivs[idx] for idx in indices),
                               ask_ivs=None if option_chain.ask_ivs is None else List(option_chain.ask_ivs[idx] for idx in indices),
                               bid_prices=None if option_chain.bid_prices is None else List(option_chain.bid_prices[idx] for idx in indices),
                               ask_prices=None if option_chain.ask_prices is None else List(option_chain.ask_prices[idx] for idx in indices))
        return option_chain

    @classmethod
    def get_uniform_chain(cls,
                          ttms: np.ndarray = np.array([0.083, 0.25]),
                          ids: np.ndarray = np.array(['1m', '3m']),
                          forwards: np.ndarray = np.array([1.0, 1.0]),
                          strikes: np.ndarray = np.linspace(0.9, 1.1, 3),
                          flat_vol: float = 0.2
                          ) -> OptionChain:
        return cls(ttms=ttms,
                   ids=ids,
                   forwards=forwards,
                   strikes_ttms=List([strikes for _ in ttms]),
                   bid_ivs=List([flat_vol*np.ones_like(strikes) for _ in ttms]),
                   ask_ivs=List([flat_vol*np.ones_like(strikes) for _ in ttms]),
                   optiontypes_ttms=List([np.where(strikes >= forward, 'C', 'P') for forward in forwards]))


@dataclass
class SwOptionChain:
    ccy: str
    ttms: np.ndarray  # swaption expiries
    tenors: np.ndarray
    ttms_ids: List[str]  # id of expiries, e.g. ['1y', '2y', '3y', '5y', '7y', '10y']
    tenors_ids: List[str]  # id of tenors, e.g. ['10y']
    forwards: List[np.ndarray]
    strikes_ttms: List[List[np.ndarray]]
    bid_ivs: List[List[np.ndarray]]
    ask_ivs: List[List[np.ndarray]]
    ticker: Optional[str] = None  # associated ticker

    def __post_init__(self):
        assert self.ttms.size == len(self.ttms_ids)
        assert self.tenors.size == len(self.tenors_ids)
        assert np.all(np.diff(self.ttms) >= 0) and np.all(self.ttms >= 0)  # check that expiries are sorted and positive
        assert np.all(np.diff(self.tenors) >= 0) and np.all(
            self.tenors >= 0)  # check that tenors are sorted and positive
        self.optiontypes_ttms = tuple([np.repeat('C', self.strikes_ttms[0][0].size) for ttm in
                                       self.ttms])  # np.where(pstrikes < 0.0, 'P', 'C').astype(str)
        # first dimension for tenors
        assert len(self.strikes_ttms) == len(self.tenors_ids)
        assert len(self.bid_ivs) == len(self.ask_ivs) == len(self.tenors_ids)
        # second dimesion for expiries
        assert len(self.strikes_ttms[0]) == len(self.ttms_ids)
        assert len(self.bid_ivs[0]) == len(self.ask_ivs[0]) == len(self.ttms_ids)
        # for fixed tenor and expiry, strikes must be 1D
        assert self.strikes_ttms[0][0].ndim == 1
        # forward must be of size # tenors x # expiries
        assert len(self.forwards) == len(self.tenors_ids) and self.forwards[0].size == len(self.ttms_ids)
        # strikes have same dimension and are consistent with vols
        for i, _ in enumerate(self.tenors_ids):
            for j, _ in enumerate(self.ttms_ids):
                assert self.strikes_ttms[i][j].size == self.strikes_ttms[0][0].size
                assert self.bid_ivs[i][j].size == self.ask_ivs[0][0].size
                assert self.strikes_ttms[i][j].size == self.bid_ivs[0][0].size

    @classmethod
    def create_swaption_chain_MF(cls,
                                 ccy: str,
                                 tenors: np.ndarray,
                                 tenors_ids: List[str],
                                 ttms: np.ndarray,
                                 ttms_ids: List[str],
                                 forwards: List[np.ndarray],
                                 strikes_ttms: List[List[np.ndarray]],
                                 ivs: List[List[np.ndarray]],
                                 ticker: str) -> SwOptionChain:

        # re-center strikes as we work with flat zero curve
        for idx_tenor, tenor in enumerate(tenors):
            for idx_ttm, ttm in enumerate(ttms):
                ts_sw = get_default_swap_term_structure(ttm, tenor)
                strikes_ttms[idx_tenor][idx_ttm] = strikes_ttms[idx_tenor][idx_ttm] - forwards[idx_tenor][idx_ttm] + swap_rate(ccy, ttm, ts_sw)
                forwards[idx_tenor][idx_ttm] = swap_rate(ccy, ttm, ts_sw)
                # print(f"idx_tenor={idx_tenor}, idx_ttm={idx_ttm}: OK")

        data = SwOptionChain(ccy=ccy,
                             ttms=ttms,
                             tenors=tenors,
                             ttms_ids=ttms_ids,
                             tenors_ids=tenors_ids,
                             forwards=forwards,
                             strikes_ttms=strikes_ttms,
                             bid_ivs=ivs,
                             ask_ivs=ivs,
                             ticker=ticker)

        return data

    def reduce_strikes(self, nb_otms: int):
        nb_strikes = int((self.strikes_ttms[0][0].size - 1) / 2)
        if nb_otms > nb_strikes:
            raise ValueError(f"number of strikes ={nb_otms} to reduce is > number of otm strikes ={nb_strikes}")
        range_ = range(nb_strikes - nb_otms, nb_strikes + nb_otms + 1)
        strikes = [[self.strikes_ttms[idx_tenor][idx_ttm][range_] for idx_ttm, _ in enumerate(self.ttms_ids)] for
                   idx_tenor, _ in enumerate(self.tenors_ids)]
        bid_ivs = [[self.bid_ivs[idx_tenor][idx_ttm][range_] for idx_ttm, _ in enumerate(self.ttms_ids)] for
                   idx_tenor, _ in enumerate(self.tenors_ids)]
        ask_ivs = [[self.ask_ivs[idx_tenor][idx_ttm][range_] for idx_ttm, _ in enumerate(self.ttms_ids)] for
                   idx_tenor, _ in enumerate(self.tenors_ids)]

        chain = SwOptionChain(ccy=self.ccy,
                              ttms=self.ttms,
                              tenors=self.tenors,
                              ttms_ids=self.ttms_ids,
                              tenors_ids=self.tenors_ids,
                              forwards=self.forwards,
                              strikes_ttms=strikes,
                              bid_ivs=bid_ivs,
                              ask_ivs=ask_ivs,
                              ticker=self.ticker)
        return chain

    def reduce_ttms(self, ttms_ids: List[str]):
        if not np.all(np.isin(ttms_ids, self.ttms_ids)):
            raise ValueError(f"Expiries to be removed not present if chain")
        idx_ttms = np.where(np.isin(self.ttms_ids, ttms_ids))[0]
        strikes = [[self.strikes_ttms[idx_tenor][idx_ttm] for idx_ttm in idx_ttms] for
                   idx_tenor, _ in enumerate(self.tenors_ids)]
        bid_ivs = [[self.bid_ivs[idx_tenor][idx_ttm] for idx_ttm in idx_ttms] for
                   idx_tenor, _ in enumerate(self.tenors_ids)]
        ask_ivs = [[self.ask_ivs[idx_tenor][idx_ttm] for idx_ttm in idx_ttms] for
                   idx_tenor, _ in enumerate(self.tenors_ids)]
        forwards = [np.array([self.forwards[idx_tenor][idx_ttm] for idx_ttm in idx_ttms]) for
                   idx_tenor, _ in enumerate(self.tenors_ids)]

        chain = SwOptionChain(ccy=self.ccy,
                              ttms=self.ttms[idx_ttms],
                              tenors=self.tenors,
                              ttms_ids=ttms_ids,
                              tenors_ids=self.tenors_ids,
                              forwards=forwards,
                              strikes_ttms=strikes,
                              bid_ivs=bid_ivs,
                              ask_ivs=ask_ivs,
                              ticker=self.ticker)
        return chain

    def reduce_tenors(self, tenors_ids: List[str]):
        if not np.all(np.isin(tenors_ids, self.tenors_ids)):
            raise ValueError(f"Tenors to be removed not present if chain")
        idx_tenors = np.where(np.isin(self.tenors_ids, tenors_ids))[0]
        strikes = [[self.strikes_ttms[idx_tenor][idx_ttm] for idx_ttm, _ in enumerate(self.ttms_ids)] for
                   idx_tenor in idx_tenors]
        bid_ivs = [[self.bid_ivs[idx_tenor][idx_ttm] for idx_ttm, _ in enumerate(self.ttms_ids)] for
                   idx_tenor in idx_tenors]
        ask_ivs = [[self.ask_ivs[idx_tenor][idx_ttm] for idx_ttm, _ in enumerate(self.ttms_ids)] for
                   idx_tenor in idx_tenors]
        forwards = [np.array([self.forwards[idx_tenor][idx_ttm] for idx_ttm, _ in enumerate(self.ttms_ids)]) for
                   idx_tenor in idx_tenors]

        chain = SwOptionChain(ccy=self.ccy,
                              ttms=self.ttms,
                              tenors=self.tenors[idx_tenors],
                              ttms_ids=self.ttms_ids,
                              tenors_ids=[self.tenors_ids[idx] for idx in idx_tenors],
                              forwards=forwards,
                              strikes_ttms=strikes,
                              bid_ivs=bid_ivs,
                              ask_ivs=ask_ivs,
                              ticker=self.ticker)
        return chain

    def get_chain_atm_vols(self) -> List[np.ndarray]:
        atm_vols = List()
        for idx_tenor, (forwards_tenor, strikes_tenor, vols_tenor) in enumerate(zip(self.forwards, self.strikes_ttms, self.get_mid_vols())):
            atm_vols_tenor = np.zeros_like(forwards_tenor)
            for idx, (forward, strikes, vol_slice) in enumerate(zip(forwards_tenor, strikes_tenor, vols_tenor)):
                atm_vols_tenor[idx] = np.interp(x=forward, xp=strikes, fp=vol_slice)
            atm_vols.append(np.array(atm_vols_tenor))
        return atm_vols

    def get_mid_vols(self) -> List[List[np.ndarray]]:
        return [[0.5 * (self.bid_ivs[idx_tenor][idx_ttm] + self.ask_ivs[idx_tenor][idx_ttm]) for
                 idx_ttm, _ in enumerate(self.ttms_ids)] for idx_tenor, _ in enumerate(self.tenors_ids)]

    # def get_chain_deltas(self) -> List[np.ndarray]:
    #     deltas_ttms = bsm.compute_bsm_deltas_ttms(ttms=self.ttms,
    #                                               forwards=self.forwards,
    #                                               strikes_ttms=self.strikes_ttms,
    #                                               optiontypes_ttms=self.optiontypes_ttms,
    #                                               vols_ttms=self.get_mid_vols())
    #     return deltas_ttms

    def get_chain_vegas(self, is_unit_ttm_vega: bool = False) -> List[List[np.ndarray]]:
        if is_unit_ttm_vega:
            ttms = np.ones_like(self.ttms)
        else:
            ttms = self.ttms
        vegas_chain = List()
        for idx_tenor, (forwards, strikes_ttms, mid_vols) in enumerate(zip(self.forwards, self.strikes_ttms, self.get_mid_vols())):
            vegas_ttms = bachel.compute_normal_vegas_ttms(ttms=ttms,
                                                          forwards=forwards,
                                                          strikes_ttms=tuple(strikes_ttms),
                                                          optiontypes_ttms=self.optiontypes_ttms,
                                                          vols_ttms=tuple(mid_vols))
            vegas_chain.append(vegas_ttms)
        return vegas_chain

    def compute_model_ivols_from_chain_data(self, model_prices: List[np.ndarray]) -> List[np.ndarray]:
        model_ivols = bsm.infer_bsm_ivols_from_model_chain_prices(ttms=self.ttms,
                                                                  forwards=self.forwards,
                                                                  discfactors=self.discfactors,
                                                                  strikes_ttms=self.strikes_ttms,
                                                                  optiontypes_ttms=self.optiontypes_ttms,
                                                                  model_prices_ttms=model_prices)
        return model_ivols

    @classmethod
    def get_slices_as_chain(cls, option_chain: SwOptionChain, ids: List[str]) -> SwOptionChain:
        indices = np.in1d(option_chain.ttms_ids, ids).nonzero()[0]
        option_chain = cls(ccy=option_chain.ccy,
                           ttms=option_chain.ttms[indices],
                           tenors=option_chain.tenors,
                           ttms_ids=option_chain.ttms_ids[indices],
                           tenors_ids=option_chain.tenors_ids,
                           forwards=[option_chain.forwards[j][indices] for j, _ in enumerate(option_chain.tenors_ids)],
                           strikes_ttms=[[option_chain.strikes_ttms[j][idx] for idx in indices] for j, _ in
                                         enumerate(option_chain.tenors_ids)],
                           ticker=option_chain.ticker,
                           bid_ivs=[[option_chain.bid_ivs[j][idx] for idx in indices] for j, _ in
                                    enumerate(option_chain.tenors_ids)],
                           ask_ivs=[[option_chain.ask_ivs[j][idx] for idx in indices] for j, _ in
                                    enumerate(option_chain.tenors_ids)])
        return option_chain

    @classmethod
    def remap_to_inc_delta(cls, vols: pd.Series) -> pd.Series:
        vols.index = [-x for x in vols.index]
        return vols


    @classmethod
    def remap_to_pc_delta(cls, inc_grid: np.ndarray) -> np.ndarray:
        put_cond = inc_grid < -0.5
        call_cond = inc_grid >= -0.5
        put_grid = -inc_grid[put_cond] - 1.0
        call_grid = -inc_grid[call_cond]
        grid = np.concatenate((put_grid, call_grid))

        return grid


@dataclass
class FutOptionChain:
    """
    container for chain data
    note we do not use chain as list of slices here
    for extensive use of numba we use List[np.ndarray] with per slice data
    """
    ccy: str
    ttms: np.ndarray
    forwards: np.ndarray
    strikes_ttms: List[np.ndarray]
    ttms_ids: Optional[np.ndarray]  # slice_t names
    ivs_call_ttms: List[np.ndarray]  # implied vol of call options on futures
    ivs_put_ttms: List[np.ndarray]  # implied vol of put options on futures
    ticker: Optional[str] = None  # associated ticker
    call_oi: List[np.ndarray] = None  # open interest of call options
    put_oi: List[np.ndarray] = None  # open interest of put options
    call_vol: List[np.ndarray] = None  # volume of call options
    put_vol: List[np.ndarray] = None  # volume of call options


    def __post_init__(self):
        assert self.ttms.size == len(self.ttms_ids)
        assert np.all(np.diff(self.ttms) >= 0) and np.all(self.ttms >= 0)  # check that expiries are sorted and positive
        self.optiontypes_ttms = tuple([np.repeat('C', self.strikes_ttms[idx_ttm].size) for idx_ttm, ttm in enumerate(self.ttms)])

        assert all(ivs_call_ttm.shape == ivs_put_ttm.shape for ivs_call_ttm, ivs_put_ttm in zip(self.ivs_call_ttms, self.ivs_put_ttms))

        assert len(self.ivs_call_ttms) == self.ttms.size
        assert self.ttms.shape == self.forwards.shape
        # strikes must be 1D
        assert all(strikes.ndim == 1 for strikes in self.strikes_ttms)
        for ivs_call_ttm, strikes_ttm in zip(self.ivs_call_ttms, self.strikes_ttms):
            assert ivs_call_ttm.shape == strikes_ttm.shape
        # put open interest call open interest must be specified (opr not specified) simulataneously
        assert (self.call_oi is not None and self.put_oi is not None) or (self.call_oi is None and self.put_oi is None)
        if self.call_oi is not None:
            for ivs_call_ttm, call_oi_ttm, put_oi_ttm in zip(self.ivs_call_ttms, self.call_oi, self.put_oi):
                assert ivs_call_ttm.shape == call_oi_ttm.shape == put_oi_ttm.shape

        assert (self.call_vol is not None and self.put_vol is not None) or (self.call_vol is None and self.put_vol is None)
        if self.call_vol is not None:
            for ivs_call_ttm, call_vol_ttm, put_vol_ttm in zip(self.ivs_call_ttms, self.call_vol, self.put_vol):
                assert ivs_call_ttm.shape == call_vol_ttm.shape == put_vol_ttm.shape

    def filter_by_oi(self,
                     max_strikes: int,
                     include_atm: bool) -> FutOptionChain:
        if self.call_oi is None:
            raise NotImplementedError(f"call/put open interest cannot be None")

        mid_idx = int(0.5 * (self.strikes_ttms[0].size - 1))

        strikes_oi_ttms = []
        ivs_call_oi_ttms = []
        ivs_put_oi_ttms = []
        call_oi_ttms = []
        put_oi_ttms = []

        for idx_ttm, ttm in enumerate(self.ttms):
            oi_ttm = self.call_oi[idx_ttm] + self.put_oi[idx_ttm]
            # top N options by open interest
            idxs = oi_ttm.argsort()[-max_strikes:][::-1]
            if include_atm and mid_idx not in idxs:
                raise ValueError(f"atm strike not found among top {max_strikes} liquid options")
            idxs_sort = np.sort(idxs)
            strikes_oi_ttm = self.strikes_ttms[idx_ttm][idxs_sort]
            ivs_call_oi_ttm = self.ivs_call_ttms[idx_ttm][idxs_sort]
            ivs_put_oi_ttm = self.ivs_put_ttms[idx_ttm][idxs_sort]
            call_oi_ttm = self.call_oi[idx_ttm][idxs_sort]
            put_oi_ttm = self.put_oi[idx_ttm][idxs_sort]

            strikes_oi_ttms.append(strikes_oi_ttm)
            ivs_call_oi_ttms.append(ivs_call_oi_ttm)
            ivs_put_oi_ttms.append(ivs_put_oi_ttm)
            call_oi_ttms.append(call_oi_ttm)
            put_oi_ttms.append(put_oi_ttm)

        fut_chain_oi = FutOptionChain(ccy=self.ccy,
                                      ttms=self.ttms,
                                      forwards=self.forwards,
                                      strikes_ttms=np.array(strikes_oi_ttms),
                                      ivs_call_ttms=np.array(ivs_call_oi_ttms),
                                      ivs_put_ttms=np.array(ivs_put_oi_ttms),
                                      ttms_ids=self.ttms_ids,
                                      call_oi=call_oi_ttms,
                                      put_oi=put_oi_ttms,
                                      ticker=self.ticker)

        return fut_chain_oi




    def get_mid_vols(self) -> List[np.ndarray]:
        return self.ivs_call_ttms

    def get_chain_vegas(self):
        optiontypes_ttms = np.repeat('C', self.strikes_ttms[0].size)

        vegas_ttms = bachel.compute_normal_vegas_ttms(ttms=self.ttms,
                                                      forwards=self.forwards,
                                                      strikes_ttms=self.strikes_ttms,
                                                      optiontypes_ttms=tuple(optiontypes_ttms),
                                                      vols_ttms=self.ivs_call_ttms)
        return vegas_ttms

    def reduce_ttms(self, ttms_ids: List[str]):
        if not np.all(np.isin(ttms_ids, self.ttms_ids)):
            raise ValueError(f"Expiries to be removed not present if chain")
        idx_ttms = np.where(np.isin(self.ttms_ids, ttms_ids))[0]
        strikes = [self.strikes_ttms[idx_ttm] for idx_ttm in idx_ttms]
        ivs_call_ttms = [self.ivs_call_ttms[idx_ttm] for idx_ttm in idx_ttms]
        ivs_put_ttms = [self.ivs_put_ttms[idx_ttm] for idx_ttm in idx_ttms]
        forwards = self.forwards[idx_ttms]

        assert self.call_oi is None and self.put_oi is None
        assert self.call_vol is None and self.put_vol is None

        chain = FutOptionChain(ccy=self.ccy,
                               ttms=self.ttms[idx_ttms],
                               forwards=forwards,
                               strikes_ttms=strikes,
                               ttms_ids=ttms_ids,
                               ivs_put_ttms=ivs_put_ttms,
                               ivs_call_ttms=ivs_call_ttms,
                               ticker=self.ticker)
        return chain


"""
run valuation for options on quadratic variance
"""
import numpy as np
import matplotlib.pyplot as plt
import stochvolmodels.data.test_option_chain as chains
from numba.typed import List
from stochvolmodels import (LogSVPricer, LogSvParams, compute_analytic_qvar, OptionChain,
                            VariableType, HestonPricer, HestonParams)

# these params are calibrated to the same BTC option chain
# v0=theta=1 to have flat vol term structure
LOGSV_BTC_PARAMS = LogSvParams(sigma0=1.0, theta=1.0, kappa1=3.1844, kappa2=3.058, beta=0.1514, volvol=1.8458)
BTC_HESTON_PARAMS = HestonParams(v0=1.0, theta=1.0, kappa=7.4565, rho=0.0919, volvol=4.0907)

ttms = {'1w': 1.0 / 52.0, '1m': 1.0 / 12.0, '3m': 0.25, '6m': 0.5}

# get test strikes for qv
option_chain = chains.get_qv_options_test_chain_data()
option_chain = OptionChain.get_slices_as_chain(option_chain, ids=list(ttms.keys()))

# compute forwards using QV model
forwards = np.array([compute_analytic_qvar(params=LOGSV_BTC_PARAMS, ttm=ttm, n_terms=4) for ttm in ttms.values()])
print(f"QV forwards = {forwards}")
# replace forwards to imply BSM vols
option_chain.forwards = forwards
# adjust strikes
option_chain.strikes_ttms = List(forward * strikes_ttm for forward, strikes_ttm in
                                 zip(option_chain.forwards, option_chain.strikes_ttms))

nb_path = 200000

# run log sv pricer
logsv_pricer = LogSVPricer()
fig1 = logsv_pricer.plot_model_ivols_vs_mc(option_chain=option_chain,
                                           params=LOGSV_BTC_PARAMS,
                                           variable_type=VariableType.Q_VAR,
                                           nb_path=nb_path)
fig1.suptitle('Implied variance skew by Log-Normal SV model')

# run Heston prices
heston_pricer = HestonPricer()
fig2 = heston_pricer.plot_model_ivols_vs_mc(option_chain=option_chain,
                                            params=BTC_HESTON_PARAMS,
                                            variable_type=VariableType.Q_VAR,
                                            nb_path=nb_path)
fig2.suptitle('Implied variance skew by Heston SV model')


plt.show()



import numpy as np
from enum import Enum

from typing import Tuple
from scipy.linalg import expm
from numba import njit
from numba import njit, objmode

# from stochvolmodels.examples.run_pricing_options_on_qvar import nb_path


# from stochvolmodels.utils.config import VariableType
# from RoughKernel import european_rule
# from stochvolmodels.pricers.logsv_pricer import LogSvParams

@njit(cache=False, fastmath=True)
def drift_ode_solve(nodes: np.ndarray, v0: np.ndarray, theta: float, lamda: float,
                    z0: np.ndarray, weight: np.ndarray, h: float):
    """

    Parameters
    ----------
    nodes : (fixed argument) exponents x_i, array of size (n,)
    v0 : (fixed argument) array of size (n, nb_path)
    theta : long-run level, scaled by lambda, scalar
    lamda : mean-reversion speed, scalar
    z0 : initial values, array of size (n, nb_path)
    weight : wieghts, array of size (n, )
    h : step size

    Returns
    -------
    Array of size (n, nb_path)
    """
    assert nodes.shape == v0.shape[:-1] == z0.shape[:-1] == weight.shape
    n = weight.size
    lamda_vec = np.repeat(lamda, n)
    theta_vec = np.repeat(theta, n)[:, None]
    diag_x = np.diag(nodes)
    I = np.identity(n)
    A = -np.outer(lamda_vec, weight) - diag_x  # matrix of size (n, n)
    b = theta_vec + diag_x @ v0  # vector of size (n, nb_path)
    with objmode(eAh='float64[:, ::1]'):
        eAh = expm(A * h)
    Dzh = eAh @ z0 + (np.linalg.inv(A) @ (eAh - I)) @ b  # vector of size (n, nb_path)

    return Dzh

@njit(cache=False, fastmath=True)
def diffus_sde_solve(y0: np.ndarray, weight: np.ndarray, volvol: float, h: float, nb_path: int,
                     z_rand: np.ndarray):
    assert y0.shape[:-1] == weight.shape and y0.shape[-1] == nb_path
    assert z_rand.shape == (nb_path,)
    n = weight.size
    weight_sum = np.sum(weight)
    volvol_ = volvol * weight_sum

    yw = np.dot(weight, y0)

    dW = z_rand * np.sqrt(h)
    Yh = yw * np.exp(-0.5 * volvol_ ** 2 * h + volvol_ * dW)

    Q = 1.0 / weight_sum * (Yh - yw)
    Yh_vec = y0.copy()
    for i in range(Yh_vec.shape[0]):
        Yh_vec[i] += Q

    return Yh_vec

@njit(cache=False, fastmath=True)
def drift_diffus_strand(nodes: np.ndarray, v0: np.ndarray, theta: float, lamda: float,
                        volvol: float, v_init: np.ndarray, weight: np.ndarray, h: float,
                        nb_path: int, z_rand: np.ndarray):
    """

    Parameters
    ----------
    nodes : exponents x_i, array of size (n,)
    v0 : (fixed argument) array of size (n, nb_path)
    theta : long-run level, scaled by lambda, scalar
    lamda : mean-reversion speed, scalar
    volvol :  total volatility. in other words, vartheta. scalar
    v_init :  initial values, array of size (n, nb_path)
    weight :  (fixed argument) wieghts, array of size (n, )
    h : time step size
    nb_path : number of paths for MC simulation
    z_rand : iid standard normals, shape (nb_path, )

    Returns
    -------

    """
    D_inn = drift_ode_solve(nodes, v0, theta, lamda, v_init, weight, 0.5 * h)
    S_inn = diffus_sde_solve(D_inn, weight, volvol, h, nb_path, z_rand)
    sol = drift_ode_solve(nodes, v0, theta, lamda, S_inn, weight, 0.5 * h)

    return sol


@njit(cache=False, fastmath=True)
def log_spot_full_solve2(nodes: np.ndarray, weight: np.ndarray,
                         v0: np.ndarray, y0: np.ndarray,
                         theta: float, lamda: float, log_s: np.ndarray, v: np.ndarray, y: np.ndarray,
                         rho: float, volvol: float, h: float, nb_path: int,
                         z0: np.ndarray, z1: np.ndarray):
    # raise ValueError
    assert nodes.shape == weight.shape and weight.ndim == 1
    assert v.shape[:-1] == weight.shape and v.shape[-1] == nb_path
    assert y.shape == (1, nb_path)
    assert log_s.shape == (1, nb_path)
    assert v0.shape[:-1] == weight.shape and v0.shape[-1] == nb_path
    assert y0.shape == (1, nb_path)
    assert z0.shape == (nb_path,) and z1.shape == (nb_path,)

    vol_h = drift_diffus_strand(nodes, v0, theta, lamda, volvol, v, weight, h, nb_path, z0)

    ww = weight[None, :]
    wlam = (weight * nodes)[None, :]
    vw = ww @ v
    volw_h = ww @ vol_h
    w_inv = 1.0 / np.sum(weight)

    c1 = 0.5
    c2 = 0.5
    rho_comp = np.sqrt(1.0 - rho * rho)

    term1 = 1.0 / volvol * ((volw_h - vw + c1 * h * wlam @ v + c2 * h * wlam @ vol_h - wlam @ v0 * h) * w_inv
                            - theta * h + lamda * c1 * h * vw + lamda * c2 * h * volw_h)
    term2 = c1 * h * np.square(vw) + c2 * h * np.square(volw_h)
    log_spot_h = log_s - 0.5 * term2 + rho * term1 + rho_comp * np.sqrt(term2) * z1

    y_h = y + 0.5 * h * (vw * vw + volw_h * volw_h)  # don't need it but keep for plotting

    return vol_h, y_h, log_spot_h

@njit(cache=False, fastmath=True)
def log_spot_full_combined(nodes: np.ndarray, weight: np.ndarray,
                           v0: np.ndarray,
                           theta: float, lamda: float, log_s0: float, v_init: np.ndarray,
                           rho: float, volvol: float, timegrid: np.ndarray, nb_path: int,
                           Z0: np.ndarray, Z1: np.ndarray):
    h = timegrid[1] - timegrid[0]
    assert np.all(np.isclose(np.diff(timegrid)[1:], h)) and timegrid[0] == 0.0
    assert Z0.shape == (timegrid.size - 1, nb_path) and Z1.shape == (timegrid.size - 1, nb_path)

    y0 = np.zeros((1, nb_path))

    vol_h = v_init.copy()
    y_h = np.zeros((1, nb_path))
    log_spot_h = np.ones((1, nb_path)) * log_s0

    # a, b, c = solve_coeffs(nodes, weight, v0[:, 0], theta, lamda, rho, volvol)

    for idx, _ in enumerate(timegrid[:-1]):
        vol_h, y_h, log_spot_h = log_spot_full_solve2(nodes, weight, v0, y0, theta, lamda,
                                                     log_spot_h, vol_h, y_h, rho, volvol, h, nb_path,
                                                      Z0[idx], Z1[idx])

    return log_spot_h, vol_h, y_h





from stochvolmodels.utils.config import VariableType

from stochvolmodels.utils.mc_payoffs import compute_mc_vars_payoff

from stochvolmodels.utils.mgf_pricer import (get_phi_grid,
                                             get_psi_grid,
                                             get_theta_grid,
                                             get_transform_var_grid,
                                             compute_integration_weights,
                                             vanilla_slice_pricer_with_mgf_grid,
                                             digital_slice_pricer_with_mgf_grid,
                                             slice_pricer_with_mgf_grid_with_gamma,
                                             slice_qvar_pricer_with_a_grid,
                                             pdf_with_mgf_grid)

from stochvolmodels.utils.funcs import (
    set_seed,
    compute_histogram_data,
    timer,
    to_flat_np_array,
    update_kwargs,
    ncdf,
    npdf,
    find_nearest
)

from stochvolmodels.pricers.analytic.bsm import (
    OptionType,
    compute_bsm_vanilla_price,
    compute_bsm_vanilla_slice_deltas,
    compute_bsm_vanilla_slice_prices,
    compute_bsm_forward_grid_prices,
    compute_bsm_vanilla_delta,
    compute_bsm_vanilla_grid_deltas,
    compute_bsm_strike_from_delta,
    compute_bsm_vanilla_deltas_ttms,
    compute_bsm_slice_vegas,
    compute_bsm_vegas_ttms,
    infer_bsm_implied_vol,
    infer_bsm_ivols_from_model_chain_prices,
    infer_bsm_ivols_from_model_slice_prices,
    infer_bsm_ivols_from_slice_prices
)

from stochvolmodels.pricers.analytic.bachelier import (
    compute_normal_delta,
    compute_normal_delta_from_lognormal_vol,
    compute_normal_delta_to_strike,
    compute_normal_deltas_ttms,
    compute_normal_price,
    compute_normal_slice_deltas,
    compute_normal_slice_prices,
    compute_normal_slice_vegas,
    compute_normal_vegas_ttms,
    infer_normal_implied_vol,
    infer_normal_ivols_from_chain_prices,
    infer_normal_ivols_from_model_slice_prices,
    infer_normal_ivols_from_slice_prices,
)

from stochvolmodels.pricers.analytic.tdist import (
    pdf_tdist,
    cdf_tdist,
    cum_mean_tdist,
    imply_drift_tdist,
    compute_default_prob_tdist,
    compute_forward_tdist,
    compute_vanilla_price_tdist,
    infer_implied_vol_tdist,
    infer_tdist_implied_vols_from_model_slice_prices
)

from stochvolmodels.pricers.logsv.affine_expansion import (
    ExpansionOrder,
    VariableType,
    compute_logsv_a_mgf_grid,
    func_a_ode_quadratic_terms,
    func_rhs,
    func_rhs_jac,
    get_expansion_n,
    get_init_conditions_a,
    solve_a_ode_grid,
    solve_analytic_ode_for_a,
    solve_analytic_ode_for_a0,
    solve_analytic_ode_grid_phi,
    solve_ode_for_a,
    compute_logsv_a_mgf_grid,
    solve_a_ode_grid,
    solve_ode_for_a,
)

from stochvolmodels.pricers.hawkes_jd_pricer import (
    HawkesJDParams,
    HawkesJDPricer
)

from stochvolmodels.pricers.heston_pricer import (
    BTC_HESTON_PARAMS,
    HestonParams,
    HestonPricer
)

from stochvolmodels.pricers.logsv_pricer import (
    LOGSV_BTC_PARAMS,
    LogSVPricer,
    LogsvModelCalibrationType,
    ConstraintsType,
    CalibrationEngine,
    get_randoms_for_chain_valuation,
    get_randoms_for_rough_vol_chain_valuation,
    logsv_mc_chain_pricer_fixed_randoms,
    rough_logsv_mc_chain_pricer_fixed_randoms
)
from stochvolmodels.pricers.logsv.logsv_params import LogSvParams

from stochvolmodels.pricers.gmm_pricer import (
    GmmParams,
    GmmPricer
)

from stochvolmodels.pricers.tdist_pricer import (
    TdistParams,
    TdistPricer
)


from stochvolmodels.data.option_chain import OptionChain, OptionSlice


from stochvolmodels.data.test_option_chain import (
    get_btc_test_chain_data,
    get_gld_test_chain_data,
    get_gld_test_chain_data_6m,
    get_qv_options_test_chain_data,
    get_spy_test_chain_data,
    get_sqqq_test_chain_data,
    get_vix_test_chain_data
)

from stochvolmodels.utils.plots import (
    align_x_limits_axs,
    align_y_limits_axs,
    create_dummy_line,
    fig_list_to_pdf,
    fig_to_pdf,
    set_legend_colors,
    get_n_sns_colors,
    map_deltas_to_str,
    model_param_ts,
    model_vols_ts,
    plot_model_risk_var,
    save_fig,
    save_figs,
    set_fig_props,
    set_subplot_border,
    set_y_limits,
    vol_slice_fit
)


from stochvolmodels.pricers.logsv.vol_moments_ode import compute_analytic_qvar




import numpy as np
import pandas as pd
import seaborn as sns
from enum import Enum
from typing import Dict, Tuple
from numba.typed import List

import stochvolmodels.pricers.analytic.bachelier as bachel
from stochvolmodels.pricers.factor_hjm.rate_logsv_params import MultiFactRateLogSvParams
from stochvolmodels.pricers.factor_hjm.rate_core import get_default_swap_term_structure
from stochvolmodels.pricers.factor_hjm.rate_logsv_pricer import simulate_logsv_MF, Measure


def do_mc_simulation(basis_type: str,
                     ccy: str,
                     ttms: np.ndarray,
                     x0: np.ndarray,
                     y0: np.ndarray,
                     I0: np.ndarray,
                     sigma0: np.ndarray,
                     params: MultiFactRateLogSvParams,
                     nb_path: int,
                     seed: int = None,
                     measure_type: Measure = Measure.RISK_NEUTRAL,
                     ts_sw: np.ndarray = None,
                     bxs: np.ndarray = None,
                     year_days: int = 360,
                     T_fwd: float = None,
                     ) -> Tuple[List[np.ndarray], List[np.ndarray], List[np.ndarray], List[np.ndarray]]:

    if basis_type != "NELSON-SIEGEL" :
        raise NotImplementedError
    x0s, y0s, I0s, sigma0s = simulate_logsv_MF(ttms=ttms,
                                               x0=x0,
                                               y0=y0,
                                               I0=I0,
                                               sigma0=sigma0,
                                               theta=params.theta,
                                               kappa1=params.kappa1,
                                               kappa2=params.kappa2,
                                               ts=params.ts,
                                               A=params.A,
                                               R=params.R,
                                               C=params.C,
                                               Omega=params.Omega,
                                               betaxs=params.beta.xs,
                                               volvolxs=params.volvol.xs,
                                               basis=params.basis,
                                               measure_type=measure_type,
                                               nb_path=nb_path,
                                               seed=seed,
                                               ccy=ccy,
                                               ts_sw=ts_sw,
                                               T_fwd=T_fwd,
                                               params0=params,
                                               bxs=bxs,
                                               year_days = year_days)

    return x0s, y0s, I0s, sigma0s


def calc_mc_vols(basis_type: str,
                 params: MultiFactRateLogSvParams,
                 ttm: float,
                 tenors: np.ndarray,
                 forwards: List[np.ndarray],
                 strikes_ttms: List[List[np.ndarray]],
                 optiontypes: np.ndarray,
                 is_annuity_measure: bool,
                 nb_path: int,
                 x0: np.ndarray = None,
                 y0: np.ndarray = None,
                 sigma0: np.ndarray = None,
                 I0: np.ndarray = None,
                 seed: int = None,
                 x_in_delta_space: bool = False) -> (List[np.ndarray], List[np.ndarray]):
    # checks
    assert len(strikes_ttms) == len(tenors)
    assert len(strikes_ttms[0]) == 1
    assert len(forwards) == len(tenors)
    for fwd in forwards:
        assert fwd.shape == (1,)

    ttms = np.array([ttm])
    # we simulate under risk-neutral measure only
    assert is_annuity_measure is False
    if x0 is None:
        x0 = np.zeros((nb_path, params.basis.get_nb_factors()))
    else:
        assert x0.shape == (nb_path, params.basis.get_nb_factors(),)
    if y0 is None:
        y0 = np.zeros((nb_path, params.basis.get_nb_aux_factors()))
    else:
        assert y0.shape == (nb_path, params.basis.get_nb_aux_factors(),)
    if sigma0 is None:
        sigma0 = np.ones((nb_path, 1))
    else:
        assert sigma0.shape == (nb_path, 1)
    if I0 is None:
        I0 = np.zeros((nb_path,))
    else:
        assert I0.shape == (nb_path,)

    ts_sws = []
    bond0s = []
    ann0s = []
    swap0s = []
    for tenor in tenors:
        ts_sw = get_default_swap_term_structure(expiry=ttm, tenor=tenor)
        ann0 = params.basis.annuity(t=ttm, ts_sw=ts_sw, x=x0, y=y0, ccy=params.ccy, m=0)[0]
        bond0 = params.basis.bond(0, ttm, x=x0, y=y0, ccy=params.ccy, m=0)[0]
        swap0 = params.basis.swap_rate(t=ttm, ts_sw=ts_sw, x=x0, y=y0, ccy=params.ccy)[0][0]
        ts_sws.append(ts_sw)
        bond0s.append(bond0)
        ann0s.append(ann0)
        swap0s.append(swap0)

    x0s, y0s, I0s, _ = do_mc_simulation(basis_type=basis_type,
                                        ccy=params.ccy,
                                        ttms=ttms,
                                        x0=x0,
                                        y0=y0,
                                        I0=I0,
                                        sigma0=sigma0,
                                        params=params,
                                        nb_path=nb_path,
                                        seed=None,
                                        measure_type=Measure.RISK_NEUTRAL)
    x0 = x0s[-1]
    y0 = y0s[-1]
    I0 = I0s[-1]
    mc_vols = List()
    mc_prices = List()
    mc_vols_ups = List()
    mc_vols_downs = List()
    std_factor = 1.96

    for idx_tenor, tenor in enumerate(tenors):
        ts_sw = ts_sws[idx_tenor]
        ann0 = ann0s[idx_tenor]
        bond0 = bond0s[idx_tenor]
        swap0 = swap0s[idx_tenor]
        strikes_ttm = strikes_ttms[idx_tenor][0]
        swap_mc, ann_mc, numer_mc = params.basis.calculate_swap_rate(ttm=ttm, x0=x0, y0=y0, I0=I0, ts_sw=ts_sw,
                                                                     ccy=params.ccy)
        # calculate option payoffs
        payoffsign = np.where(optiontypes == 'P', -1, 1).astype(float)
        option_mean = np.zeros_like(strikes_ttm)
        option_std = np.zeros_like(strikes_ttm)

        for idx, (strike, sign) in enumerate(zip(strikes_ttm, payoffsign)):
            option_mean[idx] = np.nanmean(1. / numer_mc * ann_mc * np.maximum(sign * (swap_mc - strike), 0)) / ann0 / bond0
            option_std[idx] = np.nanstd(1. / numer_mc * ann_mc * np.maximum(sign * (swap_mc - strike), 0)) / ann0 / bond0
            option_std[idx] = option_std[idx] / np.sqrt(nb_path)

        option_up = option_mean + std_factor * option_std
        option_down = np.maximum(option_mean - std_factor * option_std, 0.0)

        mc_ivols_mid = bachel.infer_normal_ivols_from_chain_prices(ttms=ttms,
                                                                   forwards=forwards,
                                                                   discfactors=np.ones_like(ttms),
                                                                   strikes_ttms=[strikes_ttm],
                                                                   optiontypes_ttms=[optiontypes],
                                                                   model_prices_ttms=[option_mean])
        mc_ivols_up = bachel.infer_normal_ivols_from_chain_prices(ttms=ttms,
                                                                  forwards=forwards,
                                                                  discfactors=np.ones_like(ttms),
                                                                  strikes_ttms=[strikes_ttm],
                                                                  optiontypes_ttms=[optiontypes],
                                                                  model_prices_ttms=[option_up])
        mc_ivols_down = bachel.infer_normal_ivols_from_chain_prices(ttms=ttms,
                                                                    forwards=forwards,
                                                                    discfactors=np.ones_like(ttms),
                                                                    strikes_ttms=[strikes_ttm],
                                                                    optiontypes_ttms=[optiontypes],
                                                                    model_prices_ttms=[option_down])

        mc_vols.append(mc_ivols_mid[0])
        mc_vols_ups.append(mc_ivols_up[0])
        mc_vols_downs.append(mc_ivols_down[0])

        mc_prices.append(option_mean)

    return mc_prices, mc_vols, mc_vols_ups, mc_vols_downs




"""
implementation of gaussian mixture pricer and calibration
"""
import numpy as np
import matplotlib.pyplot as plt
from dataclasses import dataclass
from scipy.optimize import minimize
from numba import njit
from numba.typed import List
from typing import Tuple
from enum import Enum
# project
import stochvolmodels.pricers.analytic.bsm as bsm
from stochvolmodels.utils.funcs import to_flat_np_array, timer, npdf
from stochvolmodels.pricers.model_pricer import ModelParams, ModelPricer
from stochvolmodels.utils.config import VariableType
from stochvolmodels.data.option_chain import OptionChain


@dataclass
class GmmParams(ModelParams):
    gmm_weights: np.ndarray
    gmm_mus: np.ndarray
    gmm_vols: np.ndarray
    ttm: float  # ttm is important as all params are fixed to this ttm, it is not part of calibration

    def sort_by_mus(self):
        indices = np.argsort(self.gmm_mus)
        self.gmm_weights = self.gmm_weights[indices]
        self.gmm_mus = self.gmm_mus[indices]
        self.gmm_vols = self.gmm_vols[indices]

    def get_get_avg_vol(self) -> float:
        return np.sqrt(np.sum(self.gmm_weights*np.square(self.gmm_vols)))

    def compute_state_pdfs(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        state_pdfs = np.zeros((len(x), len(self.gmm_weights)))
        agg_pdf = np.zeros_like(x)
        for idx, (gmm_weight, mu, vol) in enumerate(zip(self.gmm_weights, self.gmm_mus, self.gmm_vols)):
            state_pdf = npdf(x, mu=mu*self.ttm, vol=vol*np.sqrt(self.ttm))
            state_pdfs[:, idx] = state_pdf
            agg_pdf += gmm_weight*state_pdf
        return state_pdfs, agg_pdf

    def compute_pdf(self, x: np.ndarray):
        pdfs = np.zeros_like(x)
        for gmm_weight, mu, vol in zip(self.gmm_weights, self.gmm_mus, self.gmm_vols):
            pdfs = pdfs + gmm_weight*npdf(x, mu=mu*self.ttm, vol=vol*np.sqrt(self.ttm))
        return pdfs


class GmmPricer(ModelPricer):

    def price_chain(self, option_chain: OptionChain, params: GmmParams, **kwargs) -> np.ndarray:
        """
        implementation of generic method price_chain using heston wrapper for heston chain
        """
        model_prices_ttms = gmm_vanilla_chain_pricer(gmm_weights=params.gmm_weights,
                                                     gmm_mus=params.gmm_mus,
                                                     gmm_vols=params.gmm_vols,
                                                     ttms=option_chain.ttms,
                                                     forwards=option_chain.forwards,
                                                     strikes_ttms=option_chain.strikes_ttms,
                                                     optiontypes_ttms=option_chain.optiontypes_ttms,
                                                     discfactors=option_chain.discfactors)

        return model_prices_ttms

    def model_mc_price_chain(self, option_chain: OptionChain, params: GmmParams,
                             nb_path: int = 100000,
                             variable_type: VariableType = VariableType.LOG_RETURN,
                             **kwargs
                             ) -> (List[np.ndarray], List[np.ndarray]):
        raise NotImplementedError

    @timer
    def calibrate_model_params_to_chain_slice(self,
                                              option_chain: OptionChain,
                                              params0: GmmParams = None,
                                              is_vega_weighted: bool = True,
                                              is_unit_ttm_vega: bool = False,
                                              n_mixtures: int = 4,
                                              **kwargs
                                              ) -> GmmParams:
        """
        implementation of model calibration interface
        fit: GmmParams
        nb: always use option_chain with one slice because we need martingale condition per slice
        """

        ttms = option_chain.ttms
        if len(ttms) > 1:
            raise NotImplementedError(f"cannot calibrate to multiple slices")
        ttm = ttms[0]

        # p0 = (gmm_weights, gmm_mus, gmm_vols)
        if params0 is not None:
            p0 = np.concatenate((params0.gmm_weights, params0.gmm_mus, params0.gmm_vols))
            n_mixtures = len(params0.gmm_weights)
        else:
            gmm_weights = np.ones(n_mixtures) / n_mixtures
            gmm_mus = np.zeros(n_mixtures)
            gmm_vols = np.linspace(0.2, 1.0, n_mixtures)
            p0 = np.concatenate((gmm_weights, gmm_mus, gmm_vols))

        gmm_weights_bounds = [(0.0, 1.0)]*n_mixtures
        gmm_mus_bounds = [(-10.0, 10.0)]*n_mixtures
        gmm_vols_bounds = [(0.01, 4.0)]*n_mixtures
        bounds = np.concatenate((gmm_weights_bounds, gmm_mus_bounds, gmm_vols_bounds))

        x, y = option_chain.get_chain_data_as_xy()
        market_vols = to_flat_np_array(y)  # market mid quotes
        if is_vega_weighted:
            vegas_ttms = option_chain.get_chain_vegas(is_unit_ttm_vega=is_unit_ttm_vega)
            vegas_ttms = [vegas_ttm/sum(vegas_ttm) for vegas_ttm in vegas_ttms]
            weights = to_flat_np_array(vegas_ttms)
        else:
            weights = np.ones_like(market_vols)

        def parse_model_params(pars: np.ndarray) -> GmmParams:
            gmm_weights = pars[:n_mixtures]
            gmm_mus = pars[n_mixtures:2*n_mixtures]
            gmm_vols = pars[2*n_mixtures:]
            return GmmParams(gmm_weights=gmm_weights, gmm_mus=gmm_mus, gmm_vols=gmm_vols, ttm=ttm)

        def objective(pars: np.ndarray, args: np.ndarray) -> float:
            params = parse_model_params(pars=pars)
            model_vols = self.compute_model_ivols_for_chain(option_chain=option_chain, params=params)
            resid = np.nansum(weights * np.square(to_flat_np_array(model_vols) - market_vols))
            return resid

        def weights_sum(pars: np.ndarray) -> float:
            params = parse_model_params(pars=pars)
            return np.sum(params.gmm_weights) - 1.0

        def martingale(pars: np.ndarray) -> float:
            # we set to 1.0, mutplication with foward will be set by pricing
            params = parse_model_params(pars=pars)
            return np.sum(params.gmm_weights*np.exp((params.gmm_mus+0.5*params.gmm_vols*params.gmm_vols)*ttm)) - 1.0

        constraints = ({'type': 'eq', 'fun': weights_sum}, {'type': 'eq', 'fun': martingale})
        options = {'disp': True, 'ftol': 1e-10, 'maxiter': 500}

        res = minimize(objective, p0, args=None, method='SLSQP', constraints=constraints, bounds=bounds, options=options)
        fit_params = parse_model_params(pars=res.x)
        fit_params.sort_by_mus()

        return fit_params

    @timer
    def calibrate_model_params_to_chain(self,
                                        option_chain: OptionChain,
                                        is_vega_weighted: bool = True,
                                        is_unit_ttm_vega: bool = False,
                                        n_mixtures: int = 4,
                                        **kwargs
                                        ) -> List[str, GmmParams]:
        """
        model params are fitted per slice
        need to splic chain to slices
        """
        fit_params = {}
        params0 = None
        for ids_ in option_chain.ids:
            option_chain0 = OptionChain.get_slices_as_chain(option_chain, ids=[ids_])
            params0 = self.calibrate_model_params_to_chain_slice(option_chain=option_chain0,
                                                                 params0=params0,
                                                                 is_vega_weighted=is_vega_weighted,
                                                                 is_unit_ttm_vega=is_unit_ttm_vega,
                                                                 n_mixtures=n_mixtures,
                                                                 **kwargs)
            fit_params[ids_] = params0
        return fit_params


@njit
def compute_gmm_vanilla_price(gmm_weights: np.ndarray,
                              gmm_mus: np.ndarray,
                              gmm_vols: np.ndarray,
                              ttm: float,
                              forward: float,
                              strike: float,
                              optiontype: str,
                              discfactor: float = 1.0
                              ) -> float:
    """
    bsm deltas for strikes and vols
    """
    price = 0.0
    for gmm_weight, gmm_mu, gmm_vol in zip(gmm_weights, gmm_mus, gmm_vols):
        forward_i = forward*np.exp((gmm_mu+0.5*gmm_vol*gmm_vol)*ttm)
        # forward is vol-adjusted
        price_i = bsm.compute_bsm_vanilla_price(forward=forward_i,
                                                strike=strike,
                                                ttm=ttm,
                                                vol=gmm_vol,
                                                optiontype=optiontype,
                                                discfactor=1.0)
        price += gmm_weight * price_i
    return discfactor*price


@njit
def compute_gmm_vanilla_slice_prices(gmm_weights: np.ndarray,
                                     gmm_mus: np.ndarray,
                                     gmm_vols: np.ndarray,
                                     ttm: float,
                                     forward: float,
                                     strikes: np.ndarray,
                                     optiontypes: np.ndarray,
                                     discfactor: float = 1.0
                                     ) -> np.ndarray:
    """
    vectorised bsm deltas for array of aligned strikes, vols, and optiontypes
    """
    def f(strike: float, optiontype: str) -> float:
        return compute_gmm_vanilla_price(gmm_weights=gmm_weights,
                                         gmm_mus=gmm_mus,
                                         gmm_vols=gmm_vols,
                                         forward=forward,
                                         ttm=ttm,
                                         strike=strike,
                                         optiontype=optiontype,
                                         discfactor=discfactor)

    gmm_prices = np.zeros_like(strikes)
    for idx, (strike, optiontype) in enumerate(zip(strikes, optiontypes)):
        gmm_prices[idx] = f(strike, optiontype)
    return gmm_prices


@njit
def gmm_vanilla_chain_pricer(gmm_weights: np.ndarray,
                             gmm_mus: np.ndarray,
                             gmm_vols: np.ndarray,
                             ttms: np.ndarray,
                             forwards: np.ndarray,
                             strikes_ttms: Tuple[np.ndarray, ...],
                             optiontypes_ttms: Tuple[np.ndarray, ...],
                             discfactors: np.ndarray,
                             ) -> np.ndarray:
    """
    vectorised bsm deltas for array of aligned strikes, vols, and optiontypes
    """
    # outputs as numpy lists
    model_prices_ttms = List()
    for ttm, forward, discfactor, strikes_ttm, optiontypes_ttm in zip(ttms, forwards, discfactors, strikes_ttms,
                                                                      optiontypes_ttms):
        option_prices_ttm = compute_gmm_vanilla_slice_prices(gmm_weights=gmm_weights,
                                                             gmm_mus=gmm_mus,
                                                             gmm_vols=gmm_vols,
                                                             ttm=ttm,
                                                             forward=forward,
                                                             strikes=strikes_ttm,
                                                             optiontypes=optiontypes_ttm,
                                                             discfactor=discfactor)
        model_prices_ttms.append(option_prices_ttm)

    return model_prices_ttms


class UnitTests(Enum):
    CALIBRATOR = 1


def run_unit_test(unit_test: UnitTests):

    import seaborn as sns
    from stochvolmodels.utils import plots as plot
    from stochvolmodels.data.test_option_chain import get_btc_test_chain_data

    if unit_test == UnitTests.CALIBRATOR:
        option_chain = get_btc_test_chain_data()
        gmm_pricer = GmmPricer()
        fit_params = gmm_pricer.calibrate_model_params_to_chain(option_chain=option_chain)

        with sns.axes_style('darkgrid'):
            fig, axs = plt.subplots(2, 2, figsize=(14, 12), tight_layout=True)
            axs = plot.to_flat_list(axs)

        for idx, (key, params) in enumerate(fit_params.items()):
            print(f"{key}: {params}")
            option_chain0 = OptionChain.get_slices_as_chain(option_chain, ids=[key])
            gmm_pricer.plot_model_ivols_vs_bid_ask(option_chain=option_chain0, params=params, axs=[axs[idx]])

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.CALIBRATOR

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)
